#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/mac/OSMAC.C,v 1.1.1.1 1999/05/09 21:33:08 MJRoberts Exp $";
#endif

/*  Copyright (c) 1987, 1988 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  osmac  - Operating System dependent definitions for Macintosh
Function
  os_printf() - acts like printf()
  os_waitc()  - waits for a keystroke at the keyboard
  os_flush()  - forces flush of standard output without newline
  os_getc()   - return a single character from the keyboard
  os_gets()   - acts like gets()
  os_defext() - adds a default extension to a filename
  os_remext() - removes the extension from a filename
  os_init()   - os-dependent initialization
  os_term()   - os-dependent termination
  os_expause()- pause on exit (from ti) if this is a wacky OS (Mac, etc.)
  os_askfile()- prompt the user for a filename
  os_score()  - informs OS layer of current, maximum score (for status line)
  os_status() - sets output status:  0=normal, 1=status line, 2=room ldesc
Notes
  The operating system and compiler dependent routines have been isolated
  to this module.  This version is for Macintosh (using the Think C compiler).
Returns
  Varies by function.
Modified
  10/10/89 MJRoberts     - add history and editing functions
  06/25/89 MJRoberts     - add standard os_askfile, os_score, and os_status
  06/25/89 MJRoberts     - add MS-DOS ldesc display and status line
  05/30/89 MJRoberts     - add SUN_UNIX; add os_askfile
  10/26/88 MJRoberts     - add randomizer; default uses localtime()
  03/15/88 MJRoberts     - add support for Mac Aztec C, VAX/VMS
  02/02/88 A.Greenblatt  - Added support for Macintosh, Unix, Chipmunks
  01/02/88 MJRoberts     - created
*/

/******************************************************************************
* Operating System dependent definitions for Mac OS
******************************************************************************/

#include "os.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <limits.h>
#include <assert.h>
#ifndef _TADSRSC_
#include <FSp_fopen.h>
#endif

#include <Types.h>
#include <Files.h>
#include <MacWindows.h>
#include <StandardFile.h>
#include <Resources.h>
#include <SegLoad.h>
#include <Events.h>
#include <MacMemory.h>
#include <Folders.h>
#include <TextEdit.h>
#include <Lists.h>
#include <Processes.h>
#include <Sound.h>
#include <CodeFragments.h>

#include "lib.h"
#include "ossmac.h"


#ifndef OSANSI
void osnoreg() {}
#endif
/*void tzset() {}*/


/* extern declarations */
int topEvent(DialogPtr mydlg, ListHandle mylist);


Str255 statbuf = "\pThe Text Adventure Development System";
Str63  scorebuf;
int    status_mode;


/* open text file for writing; returns NULL on error */
osfildef *osfopwt(const char *fname, int typ)
{
        osfildef  *fp = fopen(fname, "w");
        
        track_fileopen("osfopwt", fname, fp);
        if (fp)
                os_settype(fname, typ);

        return fp;
}

/* open binary file for writing; returns NULL on error */
osfildef *osfopwb(const char *fname, int typ)
{
        osfildef  *fp = fopen(fname, "wb");
        
        track_fileopen("osfopwb", fname, fp);
        if (fp)
                os_settype(fname, typ);

        return fp;
}

/* open binary file for reading/writing; truncate; returns NULL on error */
osfildef *osfoprwtb(const char *fname, int typ)
{
        osfildef  *fp = fopen(fname, "w+b");
        
        track_fileopen("osfoprwtb", fname, fp);
        if (fp)
                os_settype(fname, typ);

        return fp;
}



void os_status(int stat)
{
    status_mode = stat;
}

int os_get_status()
{
        return status_mode;
}

void os_strsc(const char *p)
{
        static char lastbuf[80];
        
        if (p)
                strcpy(lastbuf, p);
        else
                p = lastbuf;
        
        StringCopy_pc(scorebuf, p, sizeof(scorebuf));
        macedShostat(0);
}

void os_score(int cur, int turncount)
{
        char buf[30];
        
        if ( turncount == -1 )
                os_strsc((char *)0);
        else
        {
                sprintf(buf, "%d/%d ", cur, turncount);
                os_strsc(buf);
        }
}

void os_printf(const char *fmt, ...)
{
        va_list  argptr;
        
        va_start(argptr, fmt);
        os_vprintf(fmt, argptr);
        va_end(argptr);
}

#ifdef DEBUG
static int min(int a, int b)
{
        return (a < b) ? a : b;
}
#endif

void os_vprintf(const char *fmt, va_list argptr)
{
        char    buf[512];
        int             len;
        
        len = vsprintf(buf, fmt, argptr);
        #ifdef DEBUG
        assert(len < 512);
        if (len >= 512) {
                dprintf("osvprintf: len == %d.  output == \"%.*s...\"\n",
                                len, min(strlen(buf), 64), buf);
        }
        #endif
        
        switch (status_mode) {
                case 2: /* ldesc display mode */
                        macedDbPut(buf);
                        break;
        
                case 0: /* don't allow ldesc display at all */
                        os_putln(buf);
                        break;
        
                case 1: /* status-line mode */ {
                        int   i;
                        char *p;
                        
                        /* Strip off any leading or trailing returns. */
                        for (p = buf; *p == '\n'; p++) ;
                        if ((i = strlen(p)) != 0 && p[i-1] == '\n' || p[i-1] == '\r')
                                p[--i] = '\0';
                        
                        /* If it's different, display the new status line. */
                        if (!StringEqual_pc(statbuf, p)) {
                                StringCopy_pc(statbuf, p, sizeof(statbuf));
                                macedShostat(1);
                        }
                        break;
                }
        }
}

char *strlwr(char *s)
{
        char *p;
        
        for ( p=s ; *p ; p++ ) if (isupper( *p )) *p = tolower( *p );
        return( s );
}

void os_waitc(void)
{
        /*
                If os_getc returns zero the first time, we have to call it
                again to clear out the extended key code.
        */
        int c = os_getc();
        if (!c) c = os_getc();
}

bool    use_statline, run_statline;

/*
   Check for a game to restore given as a startup parameter.
   Returns true if a game is to be restored, false otherwise.
   
   [This function is obsolete as of TADS 2.2.5 -- restoring
   games at startup is handled by the -r option instead.]
*/
int os_paramfile(char *buf)
{
        #pragma unused(buf)
        
        return false;
}

int os_init(int *argc, char *argv[], const char *prompt, char *buf, int bufsiz)
{
        /* This function is only called by tr and td.
        */
        #pragma unused(argc, argv, prompt, buf, bufsiz)
        
        extern void runtimeMenuSetup(void);
                // Actually in ossmac.c.
        
        use_statline = 1;
        run_statline = 1;
        StringCopy_pc(scorebuf, "", sizeof(scorebuf));
        runtimeMenuSetup();
        editinit();
        return 0;
}

void os_term(int rc)
{
        exit(rc);
}

int os_askfile(const char *prompt, char *fnp, int fnl, int prompt_type, int file_type)
{
        OSType          type;
        bool            writing;
        
        if (file_type == OSFTSAVE)
                type = 'TADV';
        else if (file_type == OSFTGAME)
                type = 'TADG';
        else
                type = 'TEXT';
        
        writing = (prompt_type == OS_AFP_SAVE);
                
        return os0askfile(prompt, fnp, fnl, type, writing);
}

#ifndef _TADSRSC_
/*
 *   os_exeseek:  open the application data fork.  We look up the application
 *   file with GetProcessInfo, then just open that file as a data file with
 *   fopen, returning the FILE pointer.  Within the file, there may be multiple
 *   data blocks; each data block is identified by a prefix that gives the type
 *   (a four-character code, which we compare with the caller's specified type
 *   code) and the seek offset of the next data block.  We scan the data blocks
 *   until we find the requested header.  If we don't see a valid prefix code
 *   for the whole file, we'll fall back to the old mechanism where the entire
 *   data fork was a single data block (the new typecode for this datablock is
 *   TGAM, so we'll match it only when this type is requested by the caller).
 */
osfildef *os_exeseek(const char *exename, const char *typ)
{
        #pragma unused(exename)
        
        Str255  buf;
        long    fragend = 0L;
        FILE   *fp = NULL;
        
        /* Check if we've got any fragment data.  If we do, find out how big it
                is so we can skip over it. */
        {
                CFragResource **cfrg;
                short                   resFile;
                
                cfrg = (CFragResource **) GetResource('cfrg', 0);
                if (cfrg) {
                        resFile = HomeResFile((Handle) cfrg);
                        if (resFile == gAppResFile) {
                                CFragResourceMember        *member;
                                UInt16                                  i;
                                bool                                    fragmentTooBig = false;
                                
                                HLock((Handle) cfrg);
                                for (i = (**cfrg).memberCount, member = &(**cfrg).firstMember;
                                         i > 0;
                                         i--, member = NextCFragResourceMemberPtr(member))
                                {
                                        /* If the fragment data is the entire fork, or it's bigger than we
                                           can deal with using fseek, bail out. */
                                        if (member->length == kCFragGoesToEOF ||
                                                member->offset + member->length > LONG_MAX)
                                        {
                                                fragmentTooBig = true;
                                                break;
                                        }
                                        else if (member->offset + member->length > fragend) {
                                                fragend = member->offset + member->length;
                                        }
                                }
                                HUnlock((Handle) cfrg);
                                ReleaseResource((Handle) cfrg);
                                
                                if (fragmentTooBig)
                                        return NULL;
                        }
                }
        }
        
        /* try to open the application file. */
        {
                OSErr                           err;
                ProcessSerialNumber     psn = { 0, kCurrentProcess };
                ProcessInfoRec          pi;
                FSSpec                          appFSSpec;
                
                pi.processInfoLength = sizeof(pi);
                pi.processName = NULL;
                pi.processAppSpec = &appFSSpec;
                err = GetProcessInformation(&psn, &pi);
                if (err == noErr)
                        fp = FSp_fopen(&appFSSpec, "rb");
        }
        if (fp == 0) return NULL;
        
        /* skip over any code fragment data. */
        if (fragend != 0)
                fseek(fp, fragend, SEEK_SET);
        
        /*
         *   check for the new-style prefix code - if not found, seek back
         *   to the start of the file and just return the data fork as-is
         */
        if (fread(buf, 4, 1, fp) != 1 || memcmp(buf, "MK22", 4))
        {
                if (memcmp(typ, "TGAM", 4))
                {
                        /* they don't want the game segment - fail */
                        fclose(fp);
                        return 0;
                }
                else
                {
                        /* must be an old-style game segment - return the whole data fork */
                        fseek(fp, fragend, SEEK_SET);
                        return fp;
                }
        }
        
        /* check each segment until we find the one we want */
        for (;;)
        {
                long    seekpos;
                
                /* get the header for this segment */
                if (fread(buf, 4, 1, fp) != 1
                        || fread(&seekpos, 4, 1, fp) != 1)
                        break;
                        
                /* if it matches, return it */
                if (!memcmp(buf, typ, 4))
                                return fp;
                                
                /* seek to the next position and keep going */
                fseek(fp, seekpos, SEEK_SET);
        }
        
        /* didn't find anything - return failure */
        fclose(fp);
        
        return NULL;
}
#endif


void os_settype(const char *file, int type)
{
        #if 0
        // This list is in os.h.
        #define OSFTGAME 0                               /* a game data file (.gam) */
        #define OSFTSAVE 1                                   /* a saved game (.sav) */
        #define OSFTLOG  2                               /* a transcript (log) file */
        #define OSFTSWAP 3                                             /* swap file */
        #define OSFTDATA 4      /* user data file (used with the TADS fopen() call) */
        #define OSFTCMD  5                                   /* QA command/log file */
        #define OSFTERRS 6                                    /* error message file */
        #define OSFTTEXT 7                     /* text file - used for source files */
        #define OSFTBIN  8          /* binary file of unknown type - resources, etc */
        #define OSFTPREF 9                                      /* preferences file */
        #endif
        
        #if _TADSRSC_
        const OSType    gAppCreator = 'TADR';
        #endif
        
        FInfo                   fndrinfo;
        Str255                  inf255;
        OSErr                   err;

/*      os_printf("*** About to set finder info for \"%s\"\n", file); */
        StringCopy_pc(inf255, file, sizeof(inf255));
        GetFInfo(inf255, 0, &fndrinfo);
        
        switch (type) {
                case OSFTGAME: fndrinfo.fdType = 'TADG'; break;
                case OSFTSAVE: fndrinfo.fdType = 'TADV'; break;
                case OSFTSWAP: fndrinfo.fdType = 'SWAP'; break;
                default:       fndrinfo.fdType = 'TEXT'; break;
        }
        
        switch (type) {
                case OSFTGAME: fndrinfo.fdCreator = 'TADR'; break;  // Game files always belong to the runtime.
                case OSFTLOG:
                case OSFTCMD:  fndrinfo.fdCreator = 'ttxt'; break;  // Script files belong to a text editor.
                default:       fndrinfo.fdCreator = gAppCreator;
        }

/*      os_printf("*** Setting type = %4.4s, creator = %4.4s\n", &fndrinfo.fdType, &fndrinfo.fdCreator); */
        err = SetFInfo(inf255, 0, &fndrinfo);
//      if (err)
//              os_printf("*** Warning: Return code from SetFInfo = %d\n", err);
}


// =======================================================================================
// == File iterator routines: os_find_first_file, os_find_next_file, and os_find_close.

struct find_ctx {
        CInfoPBRec              pb;
        const char         *pattern;
        long                    dirID;
        short                   index;
};

static void touch(char *state, int i, const char *pat)
{
loop:
        if (i > strlen(pat)) return;
        
        state[i] = 1;
        if (pat[i] == '*') {
                i++;
                goto loop;
        }
}

static bool match(const char *pat, ConstStr255Param str)
{
        size_t                  statelen = strlen(pat) + 1;
        char               *state = (char *) calloc(statelen, 1);
        char * const    finalstate = &state[statelen - 1];
        bool                    result;
        
        if (!state) {
                DPRINTF(("match: couldn't allocate state vector!"));
                return false;
        }
        
        touch(state, 0, pat);
        
        for (const uchar *p = &str[1], *pend = &str[StrLength(str)+1]; p != pend; p++) {
                int             i;
                
                *finalstate = 0;
                for (i = statelen - 2; i >= 0; i--) {
                        if (state[i]) {
                                if (pat[i] == '*')
                                        touch(state, i+1, pat);
                                else if (pat[i] == '?' || tolower(pat[i]) == tolower(*p))
                                        state[i] = 0, touch(state, i+1, pat);
                                else
                                        state[i] = 0;
                        }
                }
        }
        
        result = (*finalstate != 0);    
        free(state);
        
        return result;
}

void *os_find_first_file(const char *dir0, const char *pattern, char *outbuf,
                                                size_t outbufsiz, int *isdir)
{
        OSErr                   err;
        CInfoPBRec              pb;
        find_ctx           *ctx = NULL;
        
        if (*dir0 || !pattern) {
                FSSpec          dirFSS;
                Str255          dir;
                
                /* Make sure dir actually specifies a real directory. */
                StringCopy_pc(dir, dir0, sizeof(Str255));
                err = FSMakeFSSpec(0, 0, dir, &dirFSS);
                if (err != noErr)
                        return NULL;
                
                memset(&pb, 0, sizeof(pb));
                pb.hFileInfo.ioVRefNum = dirFSS.vRefNum;
                pb.hFileInfo.ioDirID = dirFSS.parID;
                pb.hFileInfo.ioNamePtr = dirFSS.name;
                pb.hFileInfo.ioFDirIndex = 0;
                err = PBGetCatInfoSync(&pb);
                if (err != noErr || (pb.hFileInfo.ioFlAttrib & ioDirMask) == 0)
                        return NULL;
                
                /* If we're still going, then dirFSS points to a real directory,
                   so allocate the ctx structure and start searching. */
                
                ctx = (find_ctx *) calloc(sizeof(find_ctx), 1);
                if (ctx) {
                        ctx->pattern = pattern;
                        ctx->index = 1;
                        ctx->dirID = pb.dirInfo.ioDrDirID;
                        ctx->pb.hFileInfo.ioVRefNum = dirFSS.vRefNum;
                        
                        return os_find_next_file(ctx, outbuf, outbufsiz, isdir);
                }
                else
                        return NULL;
        }
        else {
                /* Specified the current directory and a file path, so just get
                   the info on it. */
                
                FSSpec          fss;
                Str255          pat;
                
                assert(pattern);
                StringCopy_pc(pat, pattern, sizeof(Str255));
                err = FSMakeFSSpec(0, 0, pat, &fss);
                if (err != noErr)
                        return NULL;
                
                memset(&pb, 0, sizeof(pb));
                pb.hFileInfo.ioVRefNum = fss.vRefNum;
                pb.hFileInfo.ioDirID = fss.parID;
                pb.hFileInfo.ioNamePtr = fss.name;
                pb.hFileInfo.ioFDirIndex = 0;
                err = PBGetCatInfoSync(&pb);
                if (err != noErr)
                        return NULL;
                
                *isdir = ((pb.hFileInfo.ioFlAttrib & ioDirMask) != 0);
                StringCopy_cp(outbuf, fss.name, outbufsiz);
                
                return (void *) -1L;  // Magic value to make find_next return immediately.
        }
}

void *os_find_next_file(void *ctx0, char *outbuf, size_t outbufsiz, int *isdir)
{
        OSErr           err;
        Str255          fname;
        find_ctx   *ctx = (find_ctx *) ctx0;
        
        if (ctx0 == (void *) -1L)
                return NULL;
        
        assert(ctx != NULL);
        ctx->pb.hFileInfo.ioNamePtr = fname;
        
        while (1) {
                ctx->pb.hFileInfo.ioDirID = ctx->dirID;
                ctx->pb.hFileInfo.ioFDirIndex = ctx->index++;
                
                err = PBGetCatInfoSync(&ctx->pb);
                
                if (err != noErr) {
                        if (err != fnfErr)
                                DPRINTF(("Unexpected error %d in os_find_next_file.", err));
                        outbuf[0] = '\0';
                        os_find_close(ctx);
                        return NULL;
                }
                else if (ctx->pattern == NULL || match(ctx->pattern, fname)) {
                        *isdir = ((ctx->pb.hFileInfo.ioFlAttrib & ioDirMask) != 0);
                        StringCopy_cp(outbuf, fname, outbufsiz);
                        return ctx;
                }
                else {
                        /* Got something, but it doesn't match the pattern -- keep going. */
                }
        }
}

void os_find_close(void *ctx)
{
        if (ctx && ctx != (void *) -1L)
                free(ctx);
}

// =======================================================================================

#if 0
/* These functions are never used anywhere.  [CDN 11/26/97] */

/* check overwrite - not necessary on Mac because standard file does the work */
int os_chkovw(char *filename)
{
        return 0;
}

int os_rdexf(FILE *fp)
{
        return 0;
}
#endif

#ifndef _TADSRSC_
int (*os_exfil(const char *name))(void *)
{
        Str255  buf;
        Handle  h;
        short   curfile;
        
        StringCopy_pc(buf, name, sizeof(buf));
        
        curfile = CurResFile();
        UseResFile(gAppResFile);
        h = GetNamedResource('XFCN', buf);
        UseResFile(curfile);
        
        if (!h) return NULL;
        HLock(h);
        
        return (int (*)(void *)) *h;
}
#endif

int os_excall(int (*extfn)(void *), void *arg)
{
        return (*extfn)(arg);
}

int os_yield(void)
{
        return(topEvent(0L, 0L) == 2);
}

/* check to see if there's a break (command-period) pending */
int os_break(void)
{
        EventRecord theEvent;
        
        if (!EventAvail(keyDownMask + autoKeyMask, &theEvent))
                return(0);
        
        return (theEvent.message & charCodeMask) == '.' && (theEvent.modifiers & cmdKey) != 0;
}

#ifndef _TADSRSC_
int (*os_exfld(FILE *fp, unsigned len))(void *)
{
        #pragma unused(fp, len)
        
        return NULL;
}
#endif

/*
 *   Get the temporary file path.  Gets the path to the "temporary items"
 *   folder on the boot volume; this generally ensures that we have a
 *   place to write, even if we're running off of read-only media (such as
 *   a CD-ROM).  
 */
void os_get_tmp_path(char *buf)
{
        short   vref;
        long    parid;
        
        /* get the temporary items folder, and generate its path name */
        if (FindFolder(kOnSystemDisk, kTemporaryFolderType, kCreateFolder, &vref, &parid) != noErr
                || dirtopath(vref, parid, buf) != noErr)
        {
                /*
                 *   an error occurred - hand back an empty string, so that we
                 *   can at least try in the current folder 
                 */
                buf[0] = '\0';
        }
}


void os_gen_charmap_filename(char *filename, char *internal_id, char *argv0)
{
        #pragma unused(internal_id, argv0)
        
        filename[0] = '\0';
}


void os_advise_load_charmap(char *id, char *ldesc, char *sysinfo)
{
        #pragma unused(id, ldesc, sysinfo)
}


void os_tzset()
{
}


/*
 *   HTML 4 character set translation.  This translates to the Mac OS
 *   character set.  Note that this does not work in all display fonts.  
 */
void os_xlat_html4(unsigned int html4_char, char *result, size_t result_len)
{
        #pragma unused(result_len)
        
        /* default character to use for unknown charaters */
        #define INV_CHAR   " "
        #define NOT_USED   ""

        /* 
         *   Translation table - provides mappings for characters in the ISO
         *   Latin-1 subset of the HTML 4 character map (values 128-255).
         *   
         *   Characters marked "(approx)" are approximations where the actual
         *   desired character is not available in the Mac OS character mapping,
         *   but a reasonable approximation is used.  Characters marked "(approx
         *   unaccented)" are accented characters that are not available; these
         *   use the unaccented equivalent as an approximation, since this will
         *   presumably convey more meaning than a blank.
         *   
         *   Characters marked "(n/a)" have no equivalent (even approximating)
         *   in the Mac OS character set, and are mapped to spaces.
         *   
         *   Characters marked "(not used)" are not used by HTML '&' markups.  
         */
        static const char *xlat_tbl[] =
        {
                NOT_USED,       /* 128 (not used) */
                NOT_USED,       /* 129 (not used) */
                "",            /* 130 - sbquo */
                NOT_USED,       /* 131 (not used) */
                "",            /* 132 - bdquo */
                NOT_USED,       /* 133 (not used) */
                "",            /* 134 - dagger */
                "",            /* 135 - Dagger */
                NOT_USED,       /* 136 (not used) */
                "",            /* 137 - permil */
                NOT_USED,       /* 138 (not used) */
                "<",            /* 139 - lsaquo (approx) */
                "",            /* 140 - OElig */
                NOT_USED,       /* 141 (not used) */
                NOT_USED,       /* 142 (not used) */
                NOT_USED,       /* 143 (not used) */
                NOT_USED,       /* 144 (not used) */
                "",            /* 145 - lsquo */
                "",            /* 146 - rsquo */
                "",            /* 147 - ldquo */
                "",            /* 148 - rdquo */
                NOT_USED,       /* 149 (not used) */
                "",            /* 150 - endash */
                "",            /* 151 - emdash */
                NOT_USED,       /* 152 (not used) */
                "",            /* 153 - trade */
                NOT_USED,       /* 154 (not used) */
                ">",            /* 155 - rsaquo (approx) */
                "",            /* 156 - oelig */
                NOT_USED,       /* 157 (not used) */
                NOT_USED,       /* 158 (not used) */
                "",            /* 159 - Yuml */
                "",            /* 160 - nbsp */
                "",            /* 161 - iexcl */
                "",            /* 162 - cent */
                "",            /* 163 - pound */
                "",            /* 164 - curren */
                "",            /* 165 - yen */
                "|",            /* 166 - brvbar (approx) */
                "",            /* 167 - sect */
                "",            /* 168 - uml */
                "",            /* 169 - copy */
                "",            /* 170 - ordf */
                "",            /* 171 - laquo */
                "",            /* 172 - not */
                "-",            /* 173 - shy (approx) */
                "",            /* 174 - reg  */
                INV_CHAR,       /* 175 - macr (n/a) */
                "",            /* 176 - deg */
                "",            /* 177 - plusmn */
                INV_CHAR,       /* 178 - sup2 (n/a) */
                INV_CHAR,       /* 179 - sup3 (n/a) */
                "",            /* 180 - acute */
                "",            /* 181 - micro */
                "",            /* 182 - para */
                "",            /* 183 - middot */
                "",            /* 184 - cedil */
                INV_CHAR,       /* 185 - sup1 (n/a) */
                "",            /* 186 - ordm */
                "",            /* 187 - raquo */
                "1/4",          /* 188 - frac14 (approx) */
                "1/2",          /* 189 - frac12 (approx) */
                "3/4",          /* 190 - frac34 (approx) */
                "",            /* 191 - iquest */
                "",            /* 192 - Agrave */
                "",            /* 193 - Aacute */
                "",            /* 194 - Acirc */
                "",            /* 195 - Atilde */
                "",            /* 196 - Auml */
                "",            /* 197 - Aring */
                "",            /* 198 - AElig */
                "",            /* 199 - Ccedil */
                "",            /* 200 - Egrave */
                "",            /* 201 - Eacute */
                "",            /* 202 - Ecirc */
                "",            /* 203 - Euml */
                "",            /* 204 - Igrave */
                "",            /* 205 - Iacute */
                "",            /* 206 - Icirc */
                "",            /* 207 - Iuml */
                INV_CHAR,       /* 208 - ETH */
                "",            /* 209 - Ntilde */
                "",            /* 210 - Ograve */
                "",            /* 211 - Oacute */
                "",            /* 212 - Ocirc */
                "",            /* 213 - Otilde */
                "",            /* 214 - Ouml */
                "x",            /* 215 - times (approx) */
                "",            /* 216 - Oslash */
                "",            /* 217 - Ugrave */
                "",            /* 218 - Uacute */
                "",            /* 219 - Ucirc */
                "",            /* 220 - Uuml */
                "Y",            /* 221 - Yacute (approx unaccented) */
                INV_CHAR,       /* 222 - THORN */
                "",            /* 223 - szlig */
                "",            /* 224 - agrave */
                "",            /* 225 - aacute */
                "",            /* 226 - acirc */
                "",            /* 227 - atilde */
                "",            /* 228 - auml */
                "",            /* 229 - aring */
                "",            /* 230 - aelig */
                "",            /* 231 - ccedil */
                "",            /* 232 - egrave */
                "",            /* 233 - eacute */
                "",            /* 234 - ecirc */
                "",            /* 235 - euml */
                "",            /* 236 - igrave */
                "",            /* 237 - iacute */
                "",            /* 238 - icirc */
                "",            /* 239 - iuml */
                INV_CHAR,       /* 240 - eth */
                "",            /* 241 - ntilde */
                "",            /* 242 - ograve */
                "",            /* 243 - oacute */
                "",            /* 244 - ocirc */
                "",            /* 245 - otilde */
                "",            /* 246 - ouml */
                "",            /* 247 - divide */
                "",            /* 248 - oslash */
                "",            /* 249 - ugrave */
                "",            /* 250 - uacute */
                "",            /* 251 - ucirc */
                "",            /* 252 - uuml */
                "y",            /* 253 - yacute (approx unaccented) */
                INV_CHAR,       /* 254 - thorn */
                ""                     /* 255 - yuml */
        };

        /*
         *   Map certain extended characters into our 128-255 range.  If we
         *   don't recognize the character, return the default invalid
         *   character value.  
         */
        if (html4_char > 255)
        {
                switch(html4_char)
                {
                        case 338: html4_char = 140; break;
                        case 339: html4_char = 156; break;
                        case 376: html4_char = 159; break;
                        case 352: html4_char = 154; break;
                        case 353: html4_char = 138; break;
                        case 8211: html4_char = 150; break;
                        case 8212: html4_char = 151; break;
                        case 8216: html4_char = 145; break;
                        case 8217: html4_char = 146; break;
                        case 8218: html4_char = 130; break;
                        case 8220: html4_char = 147; break;
                        case 8221: html4_char = 148; break;
                        case 8222: html4_char = 132; break;
                        case 8224: html4_char = 134; break;
                        case 8225: html4_char = 135; break;
                        case 8240: html4_char = 137; break;
                        case 8249: html4_char = 139; break;
                        case 8250: html4_char = 155; break;
                        case 8482: html4_char = 153; break;

                        default:
                                /* unmappable character - return mid-dot character */
                                strcpy(result, INV_CHAR);
                                return;
                }
        }
        
        /* 
         *   if the character is in the regular ASCII zone, return it
         *   untranslated 
         */
        if (html4_char < 128)
        {
                result[0] = (unsigned char)html4_char;
                result[1] = '\0';
        }
        else
        {
                /* look up the character in our table and return the translation */
                strcpy(result, xlat_tbl[html4_char - 128]);
        }
}


/*
 *  Check the given file for existence.  Returns 0 if the file
 *  exists, -1 if not.
 */
int osfacc(const char *fname)
{
        FILE *fp;
        
        fp = fopen(fname, "r");
        if (fp)
        {
                fclose(fp);
                return 0;
        }
        else
                return -1;
}


char *osfgets(char *s, int n, osfildef *fp)
{
        /*
                This function is needed because the library implementation only
                interprets 0x0D as a newline, but a few places in TADS (notably
                the error file code) rely on detecting 0x0A as a newline as well.
        */
        char  c, *s0 = s;
        
        assert(n > 1);
        assert(s != NULL);
        
        /* - Handle bogus buffer sizes.
        */
        if (n <= 1) {
                if (n == 1) *s++ = '\0';  /* Terminate the string if there's room. */
                return s0;
        }
        
        /* - Read the string.
        */
        while ((c = fgetc(fp)) != EOF) {
                *s++ = c;
                if (c == '\n' || c == '\r' || --n <= 1) {
                        /* End of line found (or no more room in the buffer) -- */
                        /*   terminate the string and get out. */
                        if (n >= 1) *s++ = '\0';
                        return s0;
                }
        }
        
        /* - Got an EOF -- do the right thing.
        */
        assert(c == EOF);
        if (ferror(fp))
                return NULL;
        else {
                assert(feof(fp));
                if (s == s0)
                        /* Nothing was read. */
                        return NULL;
                else {
                        if (n >= 1) *s++ = '\0';
                        return s0;
                }
        }
}


int stricmp(const char *p, const char *q)
{
        char a, b;
        
        for (;; ++p, ++q)
        {
                a = *p; if (islower(a)) a = toupper(a);
                b = *q; if (islower(b)) b = toupper(b);
                if (!a && !b) return(0);
                if (a > b) return(1);
                if (a < b) return(-1);          
        }
}

int strnicmp(const char *p, const char *q, size_t lim)
{
        char a, b;
        
        for (;; ++p, ++q, --lim)
        {
                if (!lim) return(0);
                a = *p; if (islower(a)) a = toupper(a);
                b = *q; if (islower(b)) b = toupper(b);
                if (!a && !b) return(0);
                if (a > b) return(1);
                if (a < b) return(-1);
        }
}

int memicmp(const void *pv, const void *qv, size_t l)
{
        char a, b;
        const char *p, *q;

        for (p = (char *) pv, q = (char *) qv; l; ++p, ++q, --l)
        {
                a = *p; if (islower(a)) a = toupper(a);
                b = *q; if (islower(b)) b = toupper(b);
                if (a > b) return(1);
                if (a < b) return(-1);
        }
        return(0);
}

