/*
==========================================================================================

	File:		configs.h

	Contains:	Routines for dealing with command-line-based config files.

	Copyright:	 1998 by Christopher Nebel, all rights reserved.

	Change History (most recent first):
	
		<2>		 8/26/98	Neb		Reverse argv/argc arguments in wrtoptargv_t.
		<1>		 8/23/98	Neb		Initial version.  readconfig and writeconfig created
									based on code from os0tc_ma.c; writeoptions classes
									created based on older writeoption_as functions.

==========================================================================================
*/

#ifndef _CONFIGS_
#define _CONFIGS_

/*
** Functions for reading and writing text config files.  Give the name of a file
** to read from or write to that file, or NULL to use the preferences file for
** the component.
*/

const int LINE_BUF_SIZE = 2048;

int readconfig(const char *fname, const char *tname, const char *tname2, char *buf, int siz);
int writeconfig(const char *fname, const char *buf);


/*
** Classes for writing out command line options, either as a string or as an
** argc/argv style array.
*/

class writeoptions_t {
	protected:
		char   *fbuf;
	
	public:
		writeoptions_t(char *buf) : fbuf(buf) { }
		virtual void write(const char *format, ...) = 0;
};


class wrtoptargv_t : public writeoptions_t {
	private:
		int	   &argc;
		char  **argv;

	public:
		wrtoptargv_t(char *buf, int &c, char *v[])
			: writeoptions_t(buf), argc(c), argv(v) { }
		virtual void write(const char *format, ...);
};

class wrtoptstring_t : public writeoptions_t {
	public:
		wrtoptstring_t(char *buf)
			: writeoptions_t(buf) { }
		virtual void write(const char *format, ...);
};
	
#endif
