#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/msdos/OS0TR.C,v 1.1.1.1 1999/05/09 21:33:08 MJRoberts Exp $";
#endif

/* Copyright (c) 1992 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  os0tr.c - os mainline for tads2 run-time
Function
  invokes runtime from operating system command line
Notes
  none
Modified
  04/04/92 MJRoberts     - creation
*/

#include "os.h"
#include "std.h"
#include "ltk.h"
#include "trd.h"
#ifdef USE_GRAPH
# include <stdio.h>
# include "osspcx.h"
#endif
#ifdef USE_SOUND
# include <sbc.h>
# include <sbcmusic.h>
# include <sbcvoice.h>
#endif /* USE_SOUND */

int main(argc, argv)
int   argc;
char *argv[];
{
    int err;
    char *config_name;
#ifdef USE_GRAPH
    int          oldmode;
    int          newmode = 0x12;
    extern int   linewidth;
    extern int   os_cmdactive;
    extern int   os_defvmode;
    static struct
    {
        char *name;
        int   len;
        int   mode;                   /* video mode to set (if action == 0) */
        int   action;                        /* code for action to be taken */
    } *sw, switches[] =
    {
        { "/SVGA256", 7, 0x2e, 0 },                 /* SVGA 640 x 480 x 256 */
        { "/VGA16",   5, 0x12, 0 },                   /* VGA 640 x 480 x 16 */
        { "/EGA",     3, 0x10, 0 },                   /* EGA 640 x 350 x 16 */
        { "/NOSOUND", 7, 0x00, 1 },                       /* turn off sound */
        { "/SB_IO",   5, 0x00, 2 },             /* set soundblaster io port */
        { "/SB_INT",  6, 0x00, 3 },           /* set soundblaster interrupt */
        { 0,          0, 0x00 }
    };
    int          len;
    char        *last;
    int          gdrv, gmode;
#endif    
#ifdef USE_SOUND
    extern int os_sound;
    os_sound = 1;
#endif /* USE_SOUND */

    config_name = "config.tr";
#ifdef __DPMI16__
    ltkini(16384);
/*    config_name = "config16.tr"; */
#else
#endif

#ifdef USE_GRAPH
    /* detect best mode we can use */
    gdrv = os_disptype();
    switch (gdrv)
    {
    case 0x0303:                                /* EGA with a color monitor */
        newmode = 0x10;
        break;
        
    case 0x0505:                                /* VGA with a color monitor */
        newmode = 0x12;
        break;
        
    default:
        newmode = 0;
    }

    /* check last argument for special flags (listed above) */
check_another:
    if (argc)
    {
        char *p;
        
        last = argv[argc-1];
        len = strlen(last);
        for (p = last + len ; p > last && *(p-1) != '/' ; --p);
        if (p > last)
        {
            for (sw = switches ; sw->name ; ++sw)
            {
                if (len >= sw->len && !memicmp(p, sw->name, sw->len)
                    && *(p + sw->len) == '\0' || *(p + sw->len) == '=')
                {
                    switch(sw->action)
                    {
                    case 0:
                        newmode = sw->mode;
                        break;

                    case 1:
                        os_sound = 0;
                        break;

                    case 2:
                        if (*(p + sw->len) != '=')
                        {
                            printf("No port number specified in /SB_IO\n");
                            exit(1);
                        }
                        sscanf(p + sw->len + 1, "%x", &ct_io_addx);
                        break;

                    case 3:
                        if (*(p + sw->len) != '=')
                        {
                            printf("No interrupt specified in /SB_INT\n");
                            exit(1);
                        }
                        sscanf(p + sw->len + 1, "%x", &ct_int_num);
                        break;
                    }

                    if (len == sw->len)
                        --argc;
                    else
                        *(p-1) = '\0';
                    goto check_another;
                }
            }
        }
    }
    
    if (newmode == 0)
    {
        printf("Error: your monitor appears to be incompatible with this\
\nsoftware.  You can force this program to try to use one of its supported\
\nmodes by running the command again and putting one of these options at\
\nthe end of the command:\
\n\t/EGA        - use EGA mode (640 x 350, 16 colors)\
\n\t/VGA16      - use VGA mode (640 x 480, 16 colors)\
\n\t/SVGA256    - use Super-VGA mode (640 x 480, 256 colors)\n");
        exit(1);
    }
    
    os_defvmode = newmode;
    oldmode = ossvmode(0xffff);
    
    /* check for soundblaster and note whether it's available */
    if (os_sound && (sbc_check_card() & 6) == 6
        && sbfm_init() && ctvm_init() == 0)
        os_sound = 1;
    else
        os_sound = 0;
    
#ifdef NEVER
    /* display demo title screen */
    if (show_title)
    {
        pcxhdr  pix;
        FILE   *fp;
        char   *screens[] = { /*"demo1.pcx", "demo2.pcx",*/ "title2.pcx" };
        int     i;
        int     yofs;
        pcxrgb  pal[16];
        pcxrgb  oldpal[16];
        int     first_time = 1;
        int     last = sizeof(screens)/sizeof(screens[0]);
        
        for (i = 0, yofs = 140 ; i <= last ; ++i)
        {
            if (i < last)
            {
                fp = fopen(screens[i], "rb");
                memset(&pix, 0, sizeof(pix));
                pcxrh(&pix, fp, 0L);
                if (pcxalo(&pix))
                {
                    printf("Unable to allocate space for title page!\n");
                    exit(1);
                }
                pcxread(&pix, fp, 0L);
                fclose(fp);
                if (newmode != 46) pcxconv16(&pix, FALSE);
            }
            
            /* if something was already on the screen, wait for a key */
            if (!first_time)
            {
                os_waitc();
                if (newmode == 0x12)
                {
                    memset(pal, 0, sizeof(pal));
                    pcxfade16(&pix, oldpal, pal, 16);
                }
                ossclr(0, 0, 24, 80, 0);
            }
            else
                first_time = 0;

            if (i == last) break;

            /* we have a picture yet to display - display it */
            if (newmode == 46)
                pcxshow256(&pix, 0, yofs, 640, 480, 1);
            else
            {
                int j, k;
                
                memcpy(pal, pix.pcxpal, sizeof(pal));
                memset(pix.pcxpal, 0, sizeof(pix.pcxpal));
                pcxshow16(&pix, 0, yofs, 640, 480, 1);
                pcxfade16(&pix, pix.pcxpal, pal, 16);
            }
            pcxfree(&pix);
            memcpy(oldpal, pix.pcxpal, sizeof(oldpal));
        }
    }
#endif /* NEVER */

    os_init(&argc, argv, (char *)0, (char *)0, 0);

#ifdef NEVER
    os_inictl();
    linewidth = max_column - 1;
    if (sdesc_line != text_line) os_score(0, 0);

    ossloc(max_line, 0);
    osscsron();
    os_cmdactive = TRUE;
#endif /* NEVER */
#else /* USE_GRAPH */

    /*
     *   Scan for -plain argument, and apply if found 
     */
    {
        int i;
        for (i = 1 ; i < argc ; ++i)
        {
            if (!strcmp(argv[i], "-plain"))
                os_plain();
        }
    }

    os_init(&argc, argv, (char *)0, (char *)0, 0);

#endif /* USE_GRAPH */
    
    os_instbrk(1);
    err = os0main2(argc, argv, trdmain, "", config_name, 0);
    os_instbrk(0);
    
#ifdef USE_GRAPH
    /* return to original video mode if we entered a funny graphics mode */
    ossvmode(oldmode);

    /* turn off the mouse handler */
    asm {
        mov ax, 0x0021
        int 0x33
    }
#endif /* USE_GRAPH */

#ifdef USE_SOUND
    /* turn off soundblaster if we turned it on */
    if (os_sound)
    {
        ctvm_terminate();
        sbfm_terminate();
    }
#endif /* USE_SOUND */

    return(err);
}

