/* $Header: d:/cvsroot/tads/TADS2/msdos/OSDOSBAS.H,v 1.1.1.1 1999/05/09 21:33:08 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  osdosbas.h - DOS Base definitions
Function
  These definitions apply to DOS, Windows, and OS/2.
Notes
  
Modified
  10/17/98 MJRoberts  - Creation
*/

#ifndef OSDOSBAS_H
#define OSDOSBAS_H

#include <time.h>
#include <stdarg.h>
#include <io.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#ifdef TURBO
# include <conio.h>
#endif


/* ------------------------------------------------------------------------ */
/*
 *   osgen configuration options for this operating system.  These
 *   determine which routines from osgen are included.  Routines from
 *   osgen can be omitted by turning off various of these symbols, if the
 *   OS wants to provide its own more custom definition for the routine.  
 */
#define STD_OSCLS
#define STD_OS_HILITE
#define USEDOSCOMMON
#define USE_OVWCHK
#define USE_DOSEXT
#define USE_NULLPAUSE
#define USE_TIMERAND
#define USE_NULLSTYPE
#define USE_PATHSEARCH                  /* use search paths for tads files */
#define USE_NULL_SET_TITLE
#define USE_STDARG


/* ------------------------------------------------------------------------ */
/*
 *   Display configuration 
 */
#define USE_MORE         /* assume more-mode is desired (undef if not true) */

/* maximum width (in characters) of a line of text */
#define OS_MAXWIDTH  135

#ifdef RUNTIME
# define USE_FSDBUG                               /* full-screen debug mode */
# define USE_STATLINE
# define USE_HISTORY
# define USE_SCROLLBACK

# ifdef USE_SCROLLBACK
#  define OS_SBSTAT "(Review Mode)  \030=Up  \031=Down  PgUp=Page Up  \
PgDn=Page Down  F1=Exit"
# endif /* USE_SCROLLBACK */

# ifdef USE_HISTORY
#  define HISTBUFSIZE 4096
# endif /* USE_HISTORY */

/*
 *   Some global variables needed for console implementation 
 */
# ifdef OSGEN_INIT
#  define E
#  define I(a) =(a)
# else /* OSGEN_INIT */
#  define E extern
#  define I(a)
# endif /* OSGEN_INIT */

#  ifdef USE_LDESC
E int text_line I(12);
#  else /* USE_LDESC */
E int text_line I(1);
# endif /* USE_LDESC */

E int ldesc_line I(1), ldesc_column I(0), ldesc_color I(112);
E int ldesc_curline;

E int status_mode;
E int score_column I(70);
E int sdesc_line I(0), sdesc_column I(0), sdesc_color I(23);
E int debug_color I(0xe);

E int text_column I(0);
E int text_color I(7), text_bold_color I(15), text_normal_color I(7);
E int max_line I(24), max_column I(79);
E int text_lastcol;

# undef E
# undef I

#else /* RUNTIME */
# define USE_STDIO
# define STD_ASKFILE
# define USE_NULLINIT
# define USE_NULLSTAT
# define USE_NULLSCORE
#endif /* RUNTIME */


/* ------------------------------------------------------------------------ */
/*
 *   DOS-specific interfaces to low-level support routines 
 */
void ossgmx(int *max_line, int *max_column);
void ossclr(int top, int left, int bottom, int right, int color);
void ossdsp(int line, int column, int color, const char *msg);
int ossmon(void);
void ossloc(int line, int column);
void ossscu(int top_line, int left_column, int bottom_line,
            int right_column, int blank_color);
void ossscr(int top_line, int left_column, int bottom_line,
            int right_column, int blank_color);
void osssbini(unsigned int size);
void osssbdel(void);



/* 
 *   low-level console character reader - DOS-specific interface 
 */
char os_getch(void);

/*
 *   Get a character from the keyboard with a timeout.  The timeout is an
 *   interval in milliseconds; for example, a timeout of 1000 specifies
 *   that we should wait no more than one second for a character to be
 *   pressed.
 *   
 *   If a key is pressed before the timeout elapses, this should return
 *   OS_GETS_OK and put the character value in *ch. If the timeout elapses
 *   before a key is pressed, this returns OS_GETS_TIMEOUT.  If an error
 *   occurs, this returns OS_GETS_EOF.  
 */
int oss_getc_timeout(unsigned char *ch, unsigned long timeout);

/* status codes for timeout-based input functions */
#define OS_GETS_OK       0
#define OS_GETS_EOF      1
#define OS_GETS_TIMEOUT  2


/* ------------------------------------------------------------------------ */
/*
 *   Far pointers - any 32-bit version (DPMI32, Win32) makes no
 *   distinction between near and far pointers.  16-bit platforms
 *   distinguish near and far pointers; define the type 'osfar_t' to 'far'
 *   on those platforms.  
 */

#if defined(__DPMI32__) || defined(__WIN32__) || defined(WIN32)
#define __32BIT__
#define osfar_t
#endif

#ifndef osfar_t
#define osfar_t far
#endif


/* ------------------------------------------------------------------------ */
/*
 *   OS_LOADDS - special function qualifier needed for certain types of
 *   callback functions.
 */

/* define OS_LOADDS appropriately for the Microsoft compiler */
#ifdef MICROSOFT
# ifndef __WIN32__
#  define OS_LOADDS _loadds
# endif /* __WIN32__ */
#endif /* MICROSOFT */

/* if OS_LOADDS isn't defined, define it to nothing */
#ifndef OS_LOADDS
# define OS_LOADDS
#endif


/* ------------------------------------------------------------------------ */
/*
 *   General compiler configuration
 */

/* void a return value */
#define DISCARD (void)

/* copy a structure - dst and src are structures, not pointers */
#define OSCPYSTRUCT(dst, src) ((dst) = (src))

/* ANSI compiler */
#define OSANSI


/* ------------------------------------------------------------------------ */
/*
 *   General operating system interfaces
 */

/* allocate/release storage - malloc/free where supported */
/* void *osmalloc(size_t siz); */
/* void osfree(void *block); */

#if defined(_Windows) && !defined(__DPMI32__) && !defined(__WIN32__)

void  *ltk_alloc(size_t siz);
void   ltk_free(void *blk);
# define osmalloc(siz) ltk_alloc(siz)
# define osfree(block) ltk_free(block)
# define osrealloc(ptr, siz) ltk_realloc(ptr, siz)

#else /* _Windows */

# define osmalloc(siz) malloc(siz)
# define osfree(block) free(block)
# define osrealloc(ptr, siz) realloc(ptr, siz)

#endif /* _Windows */

/* main program exit codes */
#define OSEXSUCC 0                                 /* successful completion */
#define OSEXFAIL 1                                        /* error occurred */


/* ------------------------------------------------------------------------ */
/*
 *   File system interface 
 */

/* maximum length of a filename */
#ifdef _MAX_PATH
/* this is the Microsoft macro for the maximum length of a file path */
# define OSFNMAX  _MAX_PATH
#else
/* this is the Borland macro for the maximum length of a file path */
# define OSFNMAX  MAXPATH
#endif

/* normal path separator character */
#define OSPATHCHAR '\\'

/* alternate path separator characters */
#define OSPATHALT "/:"

/* directory separator character for PATH-style environment variables */
#define OSPATHSEP ';'                                        /* ':' on UNIX */


/* 
 *   OS file structure type.  All files are manipulated through pointers
 *   to this type. 
 */
typedef FILE osfildef;


/* open text file for reading; returns NULL on error */
/* osfildef *osfoprt(const char *fname, os_filetype_t typ); */
#define osfoprt(fname, typ) fopen(fname, "r")

/* open text file for writing; returns NULL on error */
/* osfildef *osfopwt(const char *fname, os_filetype_t typ); */
#define osfopwt(fname, typ) fopen(fname, "w")

/* open binary file for writing; returns NULL on error */
/* osfildef *osfopwb(const char *fname, os_filetype_t typ); */
#define osfopwb(fname, typ) fopen(fname, "wb")

/* open SOURCE file for reading - use appropriate text/binary mode */
/* osfildef *osfoprs(const char *fname, os_filetype_t typ); */
#define osfoprs(fname, typ) fopen(fname, "rb")

/* open binary file for reading; returns NULL on erorr */
/* osfildef *osfoprb(const char *fname, os_filetype_t typ); */
#define osfoprb(fname, typ) fopen(fname, "rb")

/* get a line of text from a text file (fgets semantics) */
/* char *osfgets(char *buf, size_t len, osfildef *fp); */
#define osfgets(buf, len, fp) fgets(buf, len, fp)

/* write a line of text to a text file (fputs semantics) */
/* void osfputs(const char *buf, osfildef *fp); */
#define osfputs(buf, fp) fputs(buf, fp)

/* open binary file for reading/writing; don't truncate */
/* osfildef *osfoprwb(const char *fname, os_filetype_t typ); */
#define osfoprwb(fname, typ) fopen(fname, "r+b")

/* open binary file for reading/writing; truncate; returns NULL on error */
/* osfildef *osfoprwtb(const char *fname, os_filetype_t typ); */
#define osfoprwtb(fname, typ) fopen(fname, "w+b")

/* write bytes to file; TRUE ==> error */
/* int osfwb(osfildef *fp, const uchar *buf, int bufl); */
#define osfwb(fp, buf, bufl) (fwrite(buf, bufl, 1, fp) != 1)

/* read bytes from file; TRUE ==> error */
/* int osfrb(osfildef *fp, uchar *buf, int bufl); */
#define osfrb(fp, buf, bufl) (fread(buf, bufl, 1, fp) != 1)

/* read bytes from file and return count; returns # bytes read, 0=error */
/* size_t osfrbc(osfildef *fp, uchar *buf, size_t bufl); */
#define osfrbc(fp, buf, bufl) (fread(buf, 1, bufl, fp))

/* get position in file */
/* long osfpos(osfildef *fp); */
#define osfpos(fp) ftell(fp)

/* seek position in file; TRUE ==> error */
/* int osfseek(osfildef *fp, long pos, int mode); */
#define osfseek(fp, pos, mode) fseek(fp, pos, mode)
#define OSFSK_SET  SEEK_SET
#define OSFSK_CUR  SEEK_CUR
#define OSFSK_END  SEEK_END

/* close a file */
/* void osfcls(osfildef *fp); */
#define osfcls(fp) fclose(fp)

/* delete a file - TRUE if error */
/* int osfdel(const char *fname); */
#define osfdel(fname) remove(fname)

/* access a file - 0 if file exists */
/* int osfacc(const char *fname) */
#define osfacc(fname) access(fname, 0)

/* get a character from a file */
/* int osfgetc(osfildef *fp); */
#define osfgetc(fp) fgetc(fp)

/*
 *   on 16-bit protected-mode DOS, we need to check manually for break, so
 *   do this in os_yield; on other DOS versions, we don't need to do
 *   anything in os_yield 
 */
#ifndef __DPMI16__
# define os_yield()  FALSE
#endif

/*
 *   Set "busy" cursor on/off.  For systems with a mouse cursor, set the
 *   cursor to an appropriate shape to notify the user that the system is
 *   busy.  Does nothing on DOS.
 */
/* void os_csr_busy(int show_as_busy); */
#define os_csr_busy(show_as_busy)



/* ------------------------------------------------------------------------ */
/*
 *   TADS 2 macro/function configuration.  Modern configurations always
 *   use the no-macro versions, so these definitions should always be set
 *   as shown below.  
 */
#define OS_MCM_NO_MACRO
#define ERR_NO_MACRO


#endif /* OSDOSBAS_H */

