#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/msdos/OSWIN.C,v 1.1.1.1 1999/05/09 21:33:08 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name

  oswin.c - os implementation for Windows
Function
  
Notes
  
Modified
  11/05/97 MJRoberts  - Creation
*/

#include <windows.h>
#include <string.h>

#include "os.h"
#include "trd.h"

/* saved game file extension */
static char default_saved_game_ext[OSFNMAX] = "SAV";

/* ------------------------------------------------------------------------ */
/* 
 *   dialog hook for standard file dialog - centers it on the screen 
 */
static UINT APIENTRY filedlg_hook(HWND dlg, UINT msg,
                                  WPARAM wpar, LPARAM lpar)
{
    /* if this is the post-initialization message, center the dialog */
    if (msg == WM_NOTIFY
        && ((NMHDR *)lpar)->code == CDN_INITDONE)
    {
        RECT deskrc;
        RECT dlgrc;
        int deskwid, deskht;
        int dlgwid, dlght;

        /* get the desktop area */
        GetWindowRect(GetDesktopWindow(), &deskrc);
        deskwid = deskrc.right - deskrc.left;
        deskht = deskrc.bottom - deskrc.top;

        /* get the dialog box area */
        GetWindowRect(((NMHDR *)lpar)->hwndFrom, &dlgrc);
        dlgwid = dlgrc.right - dlgrc.left;
        dlght = dlgrc.bottom - dlgrc.top;

        /* center the dialog on the screen */
        MoveWindow(((NMHDR *)lpar)->hwndFrom,
                   deskrc.left + (deskwid - dlgwid)/2,
                   deskrc.top + (deskht - dlght)/2, dlgwid, dlght, TRUE);

        /* message handled */
        return 1;
    }

    /* not handled */
    return 0;
}

/*
 *   Ask the user for a filename, using a system-dependent dialog or other
 *   mechanism.  Returns zero on success, non-zero on error.  
 */
int os_askfile(const char *prompt, char *fname_buf, int fname_buf_len,
               int prompt_type, int file_type)
{
    OPENFILENAME info;
    int ret;
    char filter[256];

    /* 
     *   Build the appropriate filter.  Most of the filters are
     *   boilerplate, but the "saved game" filter is special because we
     *   need to use the dynamically-selected default saved game extension
     *   for our filter.  This means that the "unknown" type is equally
     *   complicated, because it needs to include the dynamic saved game
     *   extension as well. 
     */
    if (file_type == OSFTSAVE || file_type == OSFTUNK)
    {
        static const char filter_start[] =
            "Saved Game Positions\0*.";
        static const char filter_end_sav[] =
            "All Files\0*.*\0\0";
        static const char filter_end_unk[] =
            "TADS Games\0*.GAM\0"
            "Text Files\0*.TXT\0"
            "Log Files\0*.LOG\0"
            "All Files\0*.*\0\0";
        size_t totlen;
        size_t extlen;

        /* 
         *   Build the filter string for saved game files.  Note that we
         *   need to use memcpy rather than strcpy, because we need to
         *   include a bunch of embedded null characters within the
         *   string.  First, copy the prefix up to the saved game
         *   extension...  
         */
        memcpy(filter, filter_start, sizeof(filter_start) - 1);
        totlen = sizeof(filter_start) - 1;

        /* now add the saved game extension that we're to use... */
        extlen = strlen(default_saved_game_ext) + 1;
        memcpy(filter + totlen, default_saved_game_ext, extlen);
        totlen += extlen;
        
        /* finish with the whole rest of it */
        if (file_type == OSFTSAVE)
            memcpy(filter + totlen, filter_end_sav, sizeof(filter_end_sav));
        else
            memcpy(filter + totlen, filter_end_unk, sizeof(filter_end_unk));
    }
    else
    {
        static const struct
        {
            const char *disp;
            const char *pat;
        }
        filters[] =
        {
            { "TADS Games",      "*.GAM" },                     /* OSFTGAME */
            { 0, 0 },                                  /* unused - OSFTSAVE */
            { "Transcripts",     "*.LOG" },                      /* OSFTLOG */
            { "Swap Files",      "*.DAT" },                     /* OSFTSWAP */
            { "Data Files",      "*.DAT" },                     /* OSFTDATA */
            { "Command Scripts", "*.CMD" },                      /* OSFTCMD */
            { "Message Files",   "*.MSG" },                      /* OSFTMSG */
            { "Text Files",      "*.TXT" },                     /* OSFTTEXT */
            { "Binary Files",    "*.DAT" },                      /* OSFTBIN */
            { "Character Maps",  "*.TCP" },                     /* OSFTCMAP */
            { "Preference Files","*.DAT" },                     /* OSFTPREF */
            { 0, 0 }                                    /* unused - OSFTUNK */
        };
        static const char filter_end[] = "All Files\0*.*\0\0";
        char *p;
        
        /* 
         *   We have an ordinary, static filter -- choose the appropriate
         *   filter from our list.  Start with the display name...
         */
        strcpy(filter, filters[file_type].disp);

        /* add the file filter pattern after the null character */
        p = filter + strlen(filter) + 1;
        strcpy(p, filters[file_type].pat);

        /* add the filter terminator after the null */
        p += strlen(p) + 1;
        memcpy(p, filter_end, sizeof(filter_end));
    }

    /* set up the open-file structure */
    info.lStructSize = sizeof(info);
    info.hwndOwner = 0;
    info.hInstance = 0;
    info.lpstrFilter = filter;
    info.lpstrCustomFilter = 0;
    info.nFilterIndex = 0;
    info.lpstrFile = fname_buf;
    fname_buf[0] = '\0';
    info.nMaxFile = fname_buf_len;
    info.lpstrFileTitle = 0;
    info.nMaxFileTitle = 0;
    info.lpstrInitialDir = 0;
    info.lpstrTitle = prompt;
    info.nFileOffset = 0;
    info.nFileExtension = 0;
    info.lpstrDefExt = 0;
    info.lCustData = 0;
    info.lpfnHook = filedlg_hook;
    info.Flags = OFN_ENABLEHOOK | OFN_EXPLORER;
    info.lpTemplateName = 0;

    /* display the dialog */
    if (prompt_type == OS_AFP_SAVE)
    {
        info.Flags |= OFN_NOCHANGEDIR | OFN_OVERWRITEPROMPT;
        ret = !GetSaveFileName(&info);
    }
    else
    {
        /* ask for a file to open */
        ret = !GetOpenFileName(&info);
    }

    /* return the result */
    return ret;
}

/* ------------------------------------------------------------------------ */
/*
 *   get filename from startup parameter, if possible; returns true and
 *   fills in the buffer with the parameter filename on success, false if
 *   no parameter file could be found 
 */
int os_paramfile(char *buf)
{
    /* this operation is not meaningful on windows */
    return 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Switch to a working directory 
 */
void os_set_pwd(const char *dir)
{
    SetCurrentDirectory(dir);
}

/*
 *   Switch the working directory to the path containing the given file 
 */
void os_set_pwd_file(const char *fname)
{
    char buf[OSFNMAX];
    char *p;
    char *last_sep;

    /* make a copy of the filename */
    strcpy(buf, fname);

    /* find the last path separator */
    for (p = buf, last_sep = 0 ; *p != '\0' ; ++p)
    {
        switch(*p)
        {
        case '/':
        case '\\':
        case ':':
            /* remember the location of this separator character */
            last_sep = p;
            break;
        }
    }

    /* 
     *   if we found a separator, the part up to the last separator is the
     *   directory, and the rest is the file; terminate the buffer at the
     *   last separator and switch to that directory 
     */
    if (last_sep != 0)
    {
        *last_sep = '\0';
        os_set_pwd(buf);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Service routine: set a registry key 
 */
static void set_reg_key(char *keyname, char *valname, char *val)
{
    DWORD disposition;
    HKEY  hkey;
    
    /* create or open the key */
    if (RegCreateKeyEx(HKEY_CLASSES_ROOT, keyname, 0, 0, 0, KEY_ALL_ACCESS,
                       0, &hkey, &disposition) == ERROR_SUCCESS)
    {
        /* set the value */
        RegSetValueEx(hkey, valname, 0, REG_SZ, (BYTE *)val, strlen(val) + 1);

        /* done with the key */
        RegCloseKey(hkey);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Set the saved-game extension 
 */
void os_set_save_ext(const char *ext)
{
    HKEY hkey;
    char keyname[MAX_PATH];
    
    /* copy the saved game extension into our buffer */
    strcpy(default_saved_game_ext, ext);

    /* 
     *   check to see if the registry keys for the saved game association
     *   are present 
     */
    sprintf(keyname, ".%s", ext);
    if (RegOpenKeyEx(HKEY_CLASSES_ROOT, keyname, 0, KEY_ALL_ACCESS, &hkey)
        == ERROR_SUCCESS)
    {
        /* 
         *   the suffix key exists, so assume the association is all set
         *   up -- all we need to do is close the key 
         */
        RegCloseKey(hkey);
    }
    else
    {
        char modname[MAX_PATH];
        char val[MAX_PATH];
        char classname[MAX_PATH];
        char cmd[MAX_PATH + 30];
        char *p;
        char *rootname;
        char *lastdot;

        /* 
         *   The suffix key doesn't exit, so we must create the
         *   associations.  
         */

        /*
         *   Generate a class name.  Use the name of our executable as the
         *   basis of the class name, and add ".SavedGame" to it.  Note
         *   that we must remove the directory prefix from the executable
         *   filename, since we want it to be a valid key value (i.e., no
         *   backslashes).  
         */
        GetModuleFileName(0, modname, sizeof(modname));
        for (p = rootname = modname ; *p != '\0' ; ++p)
        {
            /* 
             *   if this is a path separator, assume the root starts on
             *   the next character; if we find another path separator
             *   later we'll ignore the current one 
             */
            if (*p == '\\' || *p == '/' || *p == ':')
                rootname = p + 1;
        }

        /* generate the class name -- module name + ".SavedGame" */
        sprintf(classname, "%s.SavedGame", rootname);

        /*
         *   Set the suffix key -- this associates the suffix with the
         *   class name (for example, ".DeepSave" -> "Deep.exe.SavedGame") 
         */
        set_reg_key(keyname, "", classname);

        /*
         *   Set the class key's default value -- this gives a display
         *   name for files with our suffix.  We don't know the display
         *   name, so we'll just call these "<exe-name> Saved Game" files,
         *   with the ".EXE" removed from the executable's name.  
         */
        for (p = rootname, lastdot = p + strlen(p) ; *p != '\0' ; ++p)
        {
            /* if this is a dot, note it */
            if (*p == '.')
                lastdot = p;
        }
        sprintf(val, "\"%.*s\" Saved Game", lastdot - rootname, rootname);
        set_reg_key(classname, "", val);

        /*
         *   Now create the shell command for restoring these games.  For
         *   this, we need the name of the executable (in quotes), plus
         *   the -r option, plus the parameters ("%1"). 
         */
        sprintf(cmd, "\"%s\" -r \"%%1\"", modname);
        sprintf(keyname, "%s\\Shell\\Open\\Command", classname);
        set_reg_key(keyname, "", cmd);

        /*
         *   Set the icon for our file suffix - saved games use icon #2 in
         *   the executable.  
         */
        sprintf(cmd, "%s,2", modname);
        sprintf(keyname, "%s\\DefaultIcon", classname);
        set_reg_key(keyname, "", cmd);
    }
}


/* ------------------------------------------------------------------------ */
/*
 *   Generate a filename for a character mapping table.  On Windows, the
 *   filename is always simply "win" plus the internal ID plus ".tcp".  
 */
void os_gen_charmap_filename(char *filename, char *internal_id,
                             char *argv0)
{
    char *p;
    char *rootname;
    size_t pathlen;
    
    /* find the path prefix of the original executable filename */
    for (p = rootname = argv0 ; *p != '\0' ; ++p)
    {
        if (*p == '/' || *p == '\\' || *p == ':')
            rootname = p + 1;
    }

    /* copy the path prefix */
    pathlen = rootname - argv0;
    memcpy(filename, argv0, pathlen);

    /* if there's no trailing backslash, add one */
    if (pathlen == 0 || filename[pathlen - 1] != '\\')
        filename[pathlen++] = '\\';

    /* add "win_", plus the character set ID, plus the extension */
    strcpy(filename + pathlen, "win_");
    strcat(filename + pathlen, internal_id);
    strcat(filename + pathlen, ".tcp");
}

/* ------------------------------------------------------------------------ */
/*
 *   Get the system clock in milliseconds.  This is extremely easy for
 *   Win32, since there's an API that does exactly what we want.  
 */
long os_get_sys_clock_ms(void)
{
    return (long)GetTickCount();
}

/* ------------------------------------------------------------------------ */
/*
 *   Sleep until the specified time.  To avoid locking up our window,
 *   rather than using the system Sleep() API, we'll instead process and
 *   discard events until the delay is finished.  Fortunately, we have the
 *   handy os_get_event() routine to process events with a timeout; we'll
 *   just call this repeatedly, discarding the events, until the timeout
 *   expires.  
 */
void os_sleep_ms(long delay_in_milliseconds)
{
    long done_time;

    /* calculate when we'll be done */
    done_time = os_get_sys_clock_ms() + delay_in_milliseconds;

    /* discard events until the timeout expires */
    for (;;)
    {
        long cur_time;
        os_event_info_t info;

        /* get the current time */
        cur_time = GetTickCount();

        /* if we've reached the expiration point, delay no longer */
        if (cur_time >= done_time)
            break;

        /* 
         *   read an event, timing out after what's left of our delay
         *   interval 
         */
        os_get_event(done_time - cur_time, TRUE, &info);
    }
}

